from typing import Dict


class Logging:
    """Convenience class to control logging hooks for the Client.

    request_message: String that will be formatted with
        request_metadata and sent to the logger when a request is made.

    response_message: String that will be formatted with
        response_metadata and sent to the logger after a request is made.
    """

    def __init__(self, logger):
        self.logger = logger

        # Default logger messages
        self.request_message = (
            '\n{http_method} request to {route}'
            '\n Headers: {headers}'
            '\n Body: {data}'
            '\n Params: {params}'
        )

        self.response_message = (
            '\n{http_method} response from {route}'
            '\n Status Code {status_code}'
            '\n Body: {text}'
        )

    def log_request(self, metadata: Dict[str, str]) -> str:
        """Log request info.

        Arguments:
            metadata: The content of the metadata will be formatted into
            self.request_message.

        Returns: The formatted message.
        """
        message = self.request_message.format(**metadata)
        self.logger.info(message)
        return message

    def log_response(self, metadata: Dict[str, str]) -> str:
        """Log response info.

        Arguments:
            metadata: The content of the metadata will be formatted into
            self.response_message.

        Returns: The formatted message.
        """
        message = self.response_message.format(**metadata)
        self.logger.info(message)
        return message



import copy
from typing import Any, Dict, Mapping, Optional, TYPE_CHECKING, TypeVar, Union

import requests

import shibari

from .utils.headerdict import HeaderDict
from .utils.string_template import StringTemplate

if TYPE_CHECKING:
    from .client import Client


T = TypeVar('T', bound='Route')


class Route:
    """Representation of a single route in an API.

    Route objects should be created via Client.add_route().

    Example:
        >>> client = Client('http://foo.bar/v1')
        >>> fruits = client.add_route('fruits')

        >>> # The following are equivalent statements:
        >>> fruits.get()
        >>> client.fruits.get()

    Arguments:
        url: A string resembling a URI.
        trailing_slash: Add a trailing slash to the Route URI.

    Attributes:
        headers: Dictionary containing all the Route-level headers

    """

    rig = shibari.Rig('request')
    bind = rig.bind

    def __init__(
        self, client: 'Client', url: str, trailing_slash: bool = False,
    ):
        self.client = client
        self.trailing_slash = trailing_slash

        _url = StringTemplate(url)
        if trailing_slash:
            _url = StringTemplate(f'{url}/')

        self.url: Union[str, StringTemplate] = _url

        self.headers = HeaderDict()

        self.callables: Dict[str, Route] = {}
        self.children: Dict[str, Route] = {}

        # Parameters from parent Route
        # ie: in /foo/${barId}/${bazId}, bazId stores barId's value
        # This gets overwritten if new values are given
        self.prev_kwargs: Mapping[str, str] = {}

        self.session = self.client.new_session()
        self.session.auth = self.client.auth

    def __repr__(self):  # NOQA
        return f"Route: <{str(self.url)}>"

    def __deepcopy__(self, memodict):
        """Copy in such a way as to avoid copying the client object."""
        new = type(self)(self.client, str(self.url), self.trailing_slash)
        new.callables = copy.deepcopy(self.callables)
        new.children = copy.deepcopy(self.children)
        new.prev_kwargs = copy.deepcopy(self.prev_kwargs)

        for k, v in new.children.items():
            setattr(new, k, v)

        return new

    def __call__(self, **kwargs: str) -> T:  # NOQA C90
        """If a Route has callables, it can be called with one argument.

        This argument will be placed into the URL of the callable.

        Returns:
            Route: A Route object with the url formatted by the argument.
        """
        if len(kwargs) < 1:
            raise ValueError('Expected one keyword argument, got zero.')

        if len(kwargs) > 1:
            raise ValueError('Expected one keyword argument, got multiple.')

        # Argument has a Route associated with it
        k = list(kwargs.keys())[0]
        next_route = self.callables.get(k)

        if not next_route:
            raise ValueError(
                f'Route "{self.url}" does not take arguments',
            )

        # Ensure all known arguments are preserved.
        next_kwargs: Mapping[str, str] = {**self.prev_kwargs, **kwargs}

        # Always make a copy of the callable.
        # Else multiple calls to the same same Route will overwrite each other.
        copied_route = copy.deepcopy(next_route)
        # Update url with known arguments.
        copied_route.prev_kwargs = next_kwargs
        copied_route.url = copied_route.url.format(**next_kwargs)

        # Callable might have it's own callables.
        # Add info about used arguments
        for subcallable in copied_route.callables.values():
            subcallable.url = subcallable.url.format(**next_kwargs)

        # Children of this callable should know about the last arguments.
        for _, child_route in copied_route.children.items():
            copied_route._update_url(next_kwargs)

            for callable_route in child_route.callables.values():
                callable_route.url = callable_route.url.format(**next_kwargs)

        return copied_route

    def _update_url(self, new_kwargs: str) -> None:
        """Update this Route and it's children's urls with new values."""
        self.url = self.url.format(**new_kwargs)
        self.prev_kwargs = new_kwargs

        for _, child_route in self.children.items():
            child_route._update_url(new_kwargs)

    def post(self, *args: Any, **kwargs: Any) -> requests.Response:
        """Send a POST request."""
        return self.request('POST', self, *args, **kwargs)

    def put(self, *args: Any, **kwargs: Any) -> requests.Response:
        """Send a PUT request."""
        return self.request('PUT', self, *args, **kwargs)

    def get(self, *args: Any, **kwargs: Any) -> requests.Response:
        """Send a GET request."""
        return self.request('GET', self, *args, **kwargs)

    def delete(self, *args: Any, **kwargs: Any) -> requests.Response:
        """Send a DELETE request."""
        return self.request('DELETE', self, *args, **kwargs)

    def request(self,
                http_method: str,
                *args: Any,
                headers: Optional[Union[Dict[str, str], None]] = None,
                **kwargs: Any,
                ) -> requests.Response:
        """Send an HTTP Request.

        Accepts the same keyword arguments as requests.Session.request().

        Arguments:
            http_method: HTTP method to use for the request
        """
        self.client.rig.rigs['request'] = {}
        self.rig.rigs['request'] = {}

        local_headers = headers or {}

        request_metadata = {
            'http_method': http_method,
            'headers': local_headers,
            'route': self.url,
            'data': kwargs.get('data') or kwargs.get('json'),
            'params': kwargs.get('params'),
        }

        evaluated_headers: Dict[str, str] = {
            **self.client.headers.run_functions(self.client, request_metadata),
            **self.headers.run_functions(self, request_metadata),
            **local_headers,
        }

        request_metadata['headers'] = evaluated_headers

        for fn in self.client.hooks['request']:
            fn(request_metadata)

        evaluated_kwargs: Dict[str, Any] = {
            **kwargs,
            **self.client.request_kwargs,
        }

        response = self.session.request(
            http_method,
            self.url,
            headers=evaluated_headers,
            **evaluated_kwargs,
        )

        response_metadata = {
            'http_method': http_method,
            'route': self.url,
            'status_code': response.status_code,
            'text': response.text,
        }

        for fn in self.client.hooks['response']:
            fn(response_metadata)

        return response



import os

from setuptools import find_packages, setup


def read(filename: str):
    path = os.path.join(os.path.dirname(__file__), filename)
    with open(path, 'r') as f:
        return f.read()


def get_version_data() -> dict:
    """Read the project's version file as text."""
    data = {}

    path = os.path.join(os.path.dirname(__file__), 'inori', 'version.py')

    with open(path) as fp:
        exec(fp.read(), data)

    return data


version_data = get_version_data()


setup(
    name="inori",
    version=version_data['__version__'],
    description="The Universal API Client Constructor.",
    long_description=read('README.rst'),
    author="Joshua Fehler",
    license="GPLv3",
    url="https://github.com/jsfehler/inori",
    packages=find_packages(),
    install_requires=[
        'requests>=2.26.0',
        'shibari>=0.0.2',
    ],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ],
)



import logging
import uuid
from typing import Any, Dict, List, Union

import requests

import shibari

from .logging import Logging
from .route import Route
from .utils.headerdict import HeaderDict
from .utils.sanitize import safe_illegal_character, safe_keyword


class Client:
    """Base for an API instance.

    Client takes a string to use as the base for the API.

    Route objects are built by calling Client.add_route() with
    a string template. Route objects then become attributes of the Client.

    Example:
        >>> client = Client('http://my.service/api/v777')
        >>> client.add_route('fruits/oranges')

        >>> result_1 = client.fruits.get()
        >>> result_2 = client.fruits.oranges.get()

    Routes can use any standard HTTP verb, or make a request directly:

    Example:
        >>> client = Client('http://my.service/api/v777')
        >>> client.add_route('fruits/oranges')

        >>> result = client.fruits.request('GET')

    Routes requiring attributes are declared using a template,
    then provided by calling the Route.

    Example:

        >>> client = Client('http://my.service/api/v777')
        >>> client.add_route('fruits')
        >>> client.add_route('fruits/${fruitId}')

        >>> response = client.fruits(fruitId='8').get()

    The string template doesn't need to be provided in pieces,
    every piece will be turned into an attribute:

    Example:

        >>> client = Client('http://my.service/api/v777')
        >>> client.add_route('fruits/${fruitId}')

        >>> response = client.fruits(fruitId='5').get()

    Python keywords get an underscore prefix.

    Example:

        >>> client = Client('http://my.service/api/v777')
        >>> client.add_route('import/${importId}')

        >>> response = client._import(importId='5').get()

    Arguments:
        base_uri: Base URI for the API.

    Attributes:
        headers: Dictionary containing all Client-level headers.

        request_kwargs: Dictionary of any arguments to send with every
            request. Applies to all registered Routes.

        logger: Unique logger instance for each Client instance.

    """

    rig = shibari.Rig('request')
    bind = rig.bind

    route_paths: List[str] = []

    def __init__(self, base_uri: str, auth=None):
        self.base_uri = base_uri
        self.auth = auth

        for route in self.route_paths:
            self.add_route(route)

        self.headers = HeaderDict()

        # Keyword args that will be sent on every request made by a Route.
        self.request_kwargs: Dict[str, Any] = {}

        self.logger = logging.getLogger(f'{__name__} {str(uuid.uuid4())}')
        self.logger.addHandler(logging.NullHandler())

        self.logging = Logging(self.logger)

        self.hooks = {
            "request": [
                self.logging.log_request,
            ],
            "response": [
                self.logging.log_response,
            ],
        }

    def new_session(self) -> requests.Session:
        """Get a new instance of requests.Session.

        Route objects will call this method during init.
        """
        return requests.Session()

    def add_route(self, path: str, trailing_slash: bool = False) -> Route:
        """Take a path string and create Route objects from it.

        Route objects are automatically set as attributes of
        the Client instance.

        Arguments:
            path: URI string.
            trailing_slash: Add a trailing slash to the Route URI.

        Returns:
            The last Route that was created.
        """
        # Remove empty strings from list of pieces.
        pieces = [i for i in path.split('/') if i != '']

        # Ensure first piece is safe to use as a python variable.
        route_name: str = safe_keyword(pieces[0])
        route_name = safe_illegal_character(route_name)

        # Check if a Route already exists.
        existing_route: Union[Route, None] = getattr(self, route_name, None)

        # Create new Route if none exists
        if not existing_route:
            r = Route(
                self,
                url=f'{self.base_uri}{pieces[0]}',
                trailing_slash=trailing_slash,
            )
            setattr(self, route_name, r)

        route: Route = getattr(self, route_name)

        # Remove first piece to get list of nested paths.
        nested_pieces = pieces[1:]

        # All the found routes.
        routes = [route]
        for item in nested_pieces:
            last_route = routes[-1]

            # Identify piece

            # Callable
            if item.startswith('${') and item.endswith('}'):
                kwarg = item[2:-1]

                # Check if callable already in the last Route.
                new_route = last_route.callables.get(kwarg)
                if not new_route:
                    new_route = Route(
                        self,
                        f"{last_route.url}/{item}",
                        trailing_slash,
                    )
                    last_route.callables[kwarg] = new_route

            # Children
            else:
                nested_route_name: str = safe_keyword(item)
                nested_route_name = safe_illegal_character(nested_route_name)

                # Check if route already exists
                new_route = getattr(last_route, nested_route_name, None)
                if not new_route:
                    new_route = Route(
                        self, f"{last_route.url}/{item}",
                    )
                    last_route.children[item] = new_route
                    setattr(last_route, nested_route_name, new_route)

            routes.append(new_route)

        return routes[-1]



import keyword


def safe_keyword(word: str) -> str:
    """If word is a python keyword, add an underscore prefix.

    Arguments:
        word: A string with a single word in it.

    Returns:
        str: The word, in a safe form.
    """
    if keyword.iskeyword(word):
        return f'_{word}'
    return word


def safe_illegal_character(word: str) -> str:
    """If word has an illegal character in it, replace with an underscore.

    Arguments:
        word: A string with a single word in it.

    Returns:
        str: The word, in a safe form.
    """
    if '-' in word:
        return word.replace('-', '_')
    return word



from typing import Dict


class HeaderDict(dict):
    """Dict that can store functions via a decorator."""

    def __call__(self, key: str):
        """Decorator to store a function.

        Arguments:
            key: The name for the key to be inserted
        """
        def decorator(function):
            self[key] = function
        return decorator

    def run_functions(self, *args, **kwargs) -> Dict[str, str]:
        """Run all functions with the arguments provided."""
        rv: Dict[str, str] = {}

        for k, v in self.items():
            if callable(v):
                rv[k] = v(*args, **kwargs)
            else:
                rv[k] = v

        return rv



import string
from typing import Dict, Optional


class StringTemplate:
    """Create a string template to support partial string formatting.

    StringTemplate can compare to strings and other StringTemplate objects.

    Example:
        s = StringTemplate('${hello} ${world}')
        s.format(hello='Hello', world='World')
        'Hello World' == s
        >>> True
    """

    def __init__(self, template: str):
        self.template = string.Template(template)
        self.partial_substituted_str: str = ''

    def __eq__(self, other: object) -> bool:
        """Strings and StringTemplate can do an equality comparison."""
        return self.__repr__() == other

    def __repr__(self) -> str:
        """Return self.template.safe_substitute()."""
        return self.template.safe_substitute()

    def __str__(self) -> str:
        """Return self.template.safe_substitute()."""
        return self.__repr__()

    def format(self, mapping: Optional[Dict[str, str]]=None, **kwargs: str) -> str:  # NOQA A003
        """Format a string in such a way that a partial is allowed.

        Accepts the same arguments as string.Template.safe_substitute.

        Example:
            s = StringTemplate('${hello}' ${world}')
            s.format(world='World')
            s == '${hello} World'
            >>> True
        """
        mapping = mapping or {}

        self.partial_substituted_str = self.template.safe_substitute(
            mapping, **kwargs,
        )
        self.template = string.Template(self.partial_substituted_str)
        return self.__repr__()
